/*============================================================================
  File:     Viewing Statistics Information.sql

  Summary:  See how statistics get automatically created based on query 
			scan. View sys.indexes,	use stats_date, use dbcc show_statistics 
			and use dbcc show_statistics to produce multiple tabular data sets.
  
  Date:     November 2011

  SQL Server Version: SQL Server 2005/2008
------------------------------------------------------------------------------
  Written by Kimberly L. Tripp, SYSolutions, Inc.

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  given by Kimberly L. Tripp.  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

USE Credit
go

-- These samples use the Credit database. You can download and restore the
-- credit database from our "Past Events" page on SQLskills.com.

sp_SQLskills_SQL2008_helpindex Member
exec sp_helpstats Member
go

CREATE INDEX MemberName 
ON Member(Lastname, FirstName, MiddleInitial)
go

DBCC SHOW_STATISTICS('Member', 'MemberName')
go

-- Using Density for LastName
SELECT 10000 * 0.03846154

-- Using Density for LastName, FirstName Combo
SELECT 10000 * 0.0001

-- Is it true?
-- FOR LastName
--SELECT m.LastName, COUNT(*)
--FROM dbo.member AS m
--GROUP BY m.LastName
--go

--SELECT AVG(Counts.GroupCounts)
--FROM (SELECT m.LastName, COUNT(*) AS GroupCounts
--		FROM dbo.member AS m
--		GROUP BY m.LastName) AS Counts
--go

-- FOR LastName, FirstName
--SELECT m.LastName, m.FirstName, COUNT(*)
--FROM dbo.member AS m
--GROUP BY m.LastName, m.FirstName
--go

--SELECT AVG(Counts.GroupCounts)
--FROM (SELECT COUNT(*) AS GroupCounts
--		FROM dbo.member AS m
--		GROUP BY m.LastName, m.FirstName) AS Counts
--go

-- So - what does this tell us about the relationship between LastNames and FirstNames?
EXEC sp_helpindex member
EXEC sp_helpstats member -- look at messages window (Object does not have any stats)
go

-- If you executed the group counts, SQL Server will have created stats on firstname
-- If created, drop the stats created by prior queries
-- DROP STATISTICS member.[_WA_Sys_00000003_0CBAE877]
-- go

-- What would you expect this query to do?
SET STATISTICS IO ON --(turn on showplan with Ctrl+K)
go

SELECT m.LastName, m.FirstName, m.MiddleInitial
	, m.Phone_no, m.City
FROM dbo.Member AS m
WHERE m.FirstName LIKE 'Kim%'
go

-- Table Scan (always an option)
-- No Indexes to help find FIRSTNAMES...
-- What about scanning the NC index on LN,FN,MI and then doing bookmarks lookups...
-- seems risky? Would be good if we were guaranteed to only find a VERY small 
-- number of people with firstnames like 'Kim%'?

-- And yet it does - how did it know?

EXEC sp_helpindex member
EXEC sp_helpstats member
go

DBCC SHOW_STATISTICS('Member', '_WA_Sys_00000003_0CBAE877')
go

-- What happens if we change to 'Ki%' or 'K%'
SELECT m.LastName, m.FirstName, m.MiddleInitial
	, m.Phone_no, m.City
FROM dbo.Member AS m
WHERE m.FirstName LIKE 'Ki%'
--WHERE m.FirstName LIKE 'K%'

-- What is the selectivity of K% = SELECT 159+63*4+6+18

-- Is this always a good - generally it works well BUT nothing beats REAL statistics? 
-- In fact, SQL Server created statistics on the FirstName column...

-- Does the existence of "statistics" mean that you MUST create an index... NO but it 
-- is likely that the column is being used in SARGs and/or Join Conditions so it's 
-- something to think about - asking ITW/DTA?

-- What if the data changes:
update member	
	set firstname = 'Kimberly'
	where member_no >= 1 and member_no <= 1000
go

----------------------------------------------------

SELECT m.LastName, m.FirstName, m.MiddleInitial
	, m.Phone_no, m.City
FROM dbo.Member AS m WITH (INDEX (0))
WHERE m.FirstName LIKE 'Kim%'
go

SELECT m.LastName, m.FirstName, m.MiddleInitial
	, m.Phone_no, m.City
FROM dbo.Member AS m
WHERE m.FirstName LIKE 'Kim%'
go

----------------------------------------------------

update member	
	set firstname = 'Kimberly'
	where member_no > 1000 and member_no <= 2000
go

----------------------------------------------------

SELECT m.LastName, m.FirstName, m.MiddleInitial
	, m.Phone_no, m.City
FROM dbo.Member AS m WITH (INDEX (0))
WHERE m.FirstName LIKE 'Kim%'
go
SELECT m.LastName, m.FirstName, m.MiddleInitial
	, m.Phone_no, m.City
FROM dbo.Member AS m
WHERE m.FirstName LIKE 'Kim%'
go

-- What about seeing stats info:

-- Quick way to see stats_date for a specific table
sp_autostats 'Member'
go

-- To see all of your statistics on indexes as well as the the last time 
-- the stats were updated
SELECT 
	object_name(si.[object_id]) 	AS [TableName]
	, CASE 
		WHEN si.[index_id] = 0 then 'Heap'
		WHEN si.[index_id] = 1 then 'CL'
		WHEN si.[index_id] BETWEEN 2 AND 250 THEN 'NC ' + RIGHT('00' + convert(varchar, si.[index_id]), 3)
		ELSE ''
	  END 							AS [IndexType]
	, si.[name] 					AS [IndexName]
	, si.[index_id]					AS [IndexID]
	, CASE
		WHEN si.[index_id] BETWEEN 1 AND 250 AND STATS_DATE (si.[object_id], si.[index_id]) < DATEADD(m, -1, getdate()) 
			THEN '!! More than a month OLD !!'
		WHEN si.[index_id] BETWEEN 1 AND 250 AND STATS_DATE (si.[object_id], si.[index_id]) < DATEADD(wk, -1, getdate()) 
			THEN '! Within the past month !'
		WHEN si.[index_id] BETWEEN 1 AND 250 THEN 'Stats recent'
		ELSE ''
	  END 							AS [Warning]
	, STATS_DATE (si.[object_id], si.[index_id]) 	AS [Last Stats Update]
FROM sys.indexes AS si
WHERE OBJECTPROPERTY(si.[object_id], 'IsUserTable') = 1
ORDER BY [TableName], si.[index_id]
go

-- Or, how about ALL statistics (even those created on a col not an index)
SELECT 
	object_name(si.[object_id]) 	AS [TableName]
	, CASE 
		WHEN si.[stats_id] = 0 then 'Heap'
		WHEN si.[stats_id] = 1 then 'CL'
		WHEN INDEXPROPERTY ( si.[object_id], si.[name], 'IsAutoStatistics') = 1 THEN 'Stats-Auto'
		WHEN INDEXPROPERTY ( si.[object_id], si.[name], 'IsHypothetical') = 1 THEN 'Stats-HIND'
		WHEN INDEXPROPERTY ( si.[object_id], si.[name], 'IsStatistics') = 1 THEN 'Stats-User'
		WHEN si.[stats_id] BETWEEN 2 AND 1004 THEN 'NC ' + RIGHT('00' + convert(varchar, si.[stats_id]), 3)
		ELSE 'Text/Image'
	  END 							AS [IndexType]
	, si.[name] 					AS [IndexName]
	, si.[stats_id]					AS [IndexID]
	, CASE
		WHEN si.[stats_id] BETWEEN 1 AND 250 AND STATS_DATE (si.[object_id], si.[stats_id]) < DATEADD(m, -1, getdate()) 
			THEN '!! More than a month OLD !!'
		WHEN si.[stats_id] BETWEEN 1 AND 250 AND STATS_DATE (si.[object_id], si.[stats_id]) < DATEADD(wk, -1, getdate()) 
			THEN '! Within the past month !'
		WHEN si.[stats_id] BETWEEN 1 AND 250 THEN 'Stats recent'
		ELSE ''
	  END 							AS [Warning]
	, STATS_DATE (si.[object_id], si.[stats_id]) 	AS [Last Stats Update]
	, no_recompute
FROM sys.stats AS si
WHERE OBJECTPROPERTY(si.[object_id], 'IsUserTable') = 1
--	AND (INDEXPROPERTY ( si.[object_id], si.[name], 'IsAutoStatistics') = 1 
--			OR INDEXPROPERTY ( si.[object_id], si.[name], 'IsHypothetical') = 1 
--			OR INDEXPROPERTY ( si.[object_id], si.[name], 'IsStatistics') = 1)
ORDER BY [Last Stats Update] DESC
go


-- Seeing each tabular set from DBCC SHOW_STATISTICS 

DBCC SHOW_STATISTICS('Member', 'MemberName')
WITH STAT_HEADER 
go

DBCC SHOW_STATISTICS('Member', 'MemberName')
WITH DENSITY_VECTOR
go

DBCC SHOW_STATISTICS('Member', 'MemberName')
WITH HISTOGRAM
go

-- How do you use these programmatically
-- First, create a table into which this information will be inserted
-- Second, use dynamic string execution with an insert...exec to populate

CREATE TABLE MemberNameHistogram
(
	RANGE_HI_KEY			nvarchar(900),
	RANGE_ROWS				bigint,
	EQ_ROWS					bigint,
	DISTINCT_RANGE_ROWS		bigint,
	AVG_RANGE_ROWS			bigint,
)
go

INSERT MemberNameHistogram
EXEC ('DBCC SHOW_STATISTICS(''Member'', ''MemberName'') WITH HISTOGRAM')
go

SELECT * FROM MemberNameHistogram
go

sp_updatestats
go

sp_createstats 'indexonly', 'fullscan'
go

sp_helpstats member
go
-- New stats are listed as User-Stats

-- only "lightly" doc'ed...
DBCC SHOW_STATISTICS('Member', 'MemberName')
WITH stats_stream
go